# _version.py generated by setuptools_scm
from ._version import __version__  # type: ignore

from .arch import IterationModule, IterationLossModule
from .arch import SimilarityMatching, MultiSimilarityMatching
from .arch import SupervisedSimilarityMatching
from .util import extract_embeddings

__all__ = [
    "IterationModule",
    "IterationLossModule",
    "SimilarityMatching",
    "MultiSimilarityMatching",
    "SupervisedSimilarityMatching",
    "extract_embeddings",
    "__version__",
]
